/*
  --------------------------------------------------------------------------------------------------------------------
   Example sketch/program showing An Arduino Door Access Control featuring RFID, EEPROM, Relay
  ---------------------------------------------------------------------------------------------------------------------

  This example showing a complete Door Access Control System

  Simple Work Flow (not limited to) :
                                     +---------+
  +----------------------------------->READ TAGS+^------------------------------------------+
  |                              +--------------------+                                     |
  |                              |                    |                                     |
  |                              |                    |                                     |
  |                         +----v-----+        +-----v----+                                |
  |                         |MASTER TAG|        |OTHER TAGS|                                |
  |                         +--+-------+        ++-------------+                            |
  |                            |                 |             |                            |
  |                            |                 |             |                            |
  |                      +-----v---+        +----v----+   +----v------+                     |
  |         +------------+READ TAGS+---+    |KNOWN TAG|   |UNKNOWN TAG|                     |
  |         |            +-+-------+   |    +-----------+ +------------------+              |
  |         |              |           |                |                    |              |
  |    +----v-----+   +----v----+   +--v--------+     +-v----------+  +------v----+         |
  |    |MASTER TAG|   |KNOWN TAG|   |UNKNOWN TAG|     |GRANT ACCESS|  |DENY ACCESS|         |
  |    +----------+   +---+-----+   +-----+-----+     +-----+------+  +-----+-----+         |
  |                       |               |                 |               |               |
  |       +----+     +----v------+     +--v---+             |               +--------------->
  +-------+EXIT|     |DELETE FROM|     |ADD TO|             |                               |
          +----+     |  EEPROM   |     |EEPROM|             |                               |
                     +-----------+     +------+             +-------------------------------+

   Use a Master Card which is act as Programmer then you can able to choose card holders who will granted access or not

 * **Easy User Interface**

   Just one RFID tag needed whether Delete or Add Tags. You can choose to use Leds for output or Serial LCD module to inform users.

 * **Stores Information on EEPROM**

   Information stored on non volatile Arduino's EEPROM memory to preserve Users' tag and Master Card. No Information lost
   if power lost. EEPROM has unlimited Read cycle but roughly 100,000 limited Write cycle.

    Typical pin layout used:
   -----------------------------------------------------------------------------------------
               MFRC522      Arduino       Arduino   Arduino    Arduino          Arduino
               Reader/PCD   Uno/101       Mega      Nano v3    Leonardo/Micro   Pro Micro
   Signal      Pin          Pin           Pin       Pin        Pin              Pin
   -----------------------------------------------------------------------------------------
   RST/Reset   RST          9             5         D9         RESET/ICSP-5     RST
   SPI SS      SDA(SS)      10            53        D10        10               10
   SPI MOSI    MOSI         11 / ICSP-4   51        D11        ICSP-4           16
   SPI MISO    MISO         12 / ICSP-1   50        D12        ICSP-1           14
   SPI SCK     SCK          13 / ICSP-3   52        D13        ICSP-3           15

*/

#include <EEPROM.h>   // We are going to read and write cards IDs from/to EEPROM
#include <SPI.h>      // RC522 Module uses SPI protocol
#include <MFRC522.h>  // Library for Mifare RC522 Devices

#define SS_PIN 10
#define RST_PIN 9
MFRC522 mfrc522(SS_PIN, RST_PIN);  // Create MFRC522 instance.

#define RedLED 6
#define YellowLED 5
#define BlueLED 3

#define Relay A0
#define Taster 4
#define Buzzer 2
#define wipeB 4  // Button pin for WipeMode
boolean wipeTaster, pwipeTaster;



//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//Promjenljive i konstante za praćenje prisustva master kartice na citacu

//Potrebno vrijeme za prevođenje u matster mod
//Potrebno vrijeme za bristanje EEPROM-a
//Promjenljiva koja detektuje pristustvo master kartice
//Pomjenljiva koja bilježi trenutak detekcije master kartice
//Poromjenljiva koja indicira da treba preći u master mod ako se master kartica zadrži dovoljno dugo
//Poromjenljiva koja indicira da treba brisati EEPROM ako se master kartica zadrži dovoljno dugo



#define NO_ACTION 0
#define ACCESS_GRANTED 1
#define ACCESS_DENIDED 2
#define PROGRAM_MODE 3
#define CARD_DELETED 4
#define CARD_ADDED 5
#define NEW_MASTER 6

#define MIN_PAUZA 300
#define SIG_DURATION 3000
#define SET_LED_TIME 250

#include "pitches.h"

// notes in the melody:
int AGmelody[] = {
  NOTE_C4, NOTE_G3, NOTE_G3, NOTE_A3, NOTE_G3, 0, NOTE_B3, NOTE_C4
};
int ADmelody[] = {
  NOTE_F6, NOTE_F6, NOTE_F6, NOTE_F6, NOTE_F6, NOTE_F6, NOTE_F6, NOTE_F6
};

// note durations: 4 = quarter note, 8 = eighth note, etc.:
int AGnoteDurations[] = {
  4, 8, 8, 4, 4, 4, 4, 4
};
int ADnoteDurations[] = {
  4, 4, 4, 4, 4, 4, 4, 4
};
#define FULL_MELODY 8
#define HALF_MELODY 4
#define ONE_NOTE 1
unsigned char notes, currentNote;
int* melody;
int* noteDur;


boolean newCard;
int SigMod;
long detectionTime;
long setLED;

uint8_t successRead;  // Variable integer to keep if we have Successful Read from Reader

byte storedCard[8];  // Stores an ID read from EEPROM
byte readCard[8];    // Stores scanned ID read from RFID Module
byte masterCard[8];  //{0x3B, 0xD2, 0x2B, 0x1B}; // Master card's ID

boolean newMaster = false;

bool programMode = false;  // initialize programming mode

//Melody function
void Sound(int* melody, int* noteDurations, unsigned char curNote) {
  // To calculate the note duration, take one second divided by the note type.
  //e.g. quarter note = 1000 / 4, eighth note = 1000/8, etc.
  int noteDuration = 1000 / noteDurations[curNote];
  tone(Buzzer, melody[curNote], noteDuration);
  delay(noteDuration);
}

//Light signalisation
void Signals(int Mod) {
  static int noAct;

  if (Mod == NO_ACTION) {
    if (millis() - setLED > SET_LED_TIME) {
      if (noAct % 5) {
        digitalWrite(RedLED, LOW);
        digitalWrite(BlueLED, LOW);
        digitalWrite(YellowLED, LOW);
      } else if (noAct % 5 == 0) {
        digitalWrite(RedLED, LOW);
        digitalWrite(BlueLED, LOW);
        digitalWrite(YellowLED, HIGH);
      }
      noAct++;
      setLED = millis();
      digitalWrite(Relay, LOW);
      noTone(Buzzer);
    }
  } else if (Mod == PROGRAM_MODE) {
    if (millis() - setLED > SET_LED_TIME) {
      if (!(noAct % 5)) {
        digitalWrite(RedLED, LOW);
        digitalWrite(BlueLED, LOW);
        digitalWrite(YellowLED, LOW);
      } else if (noAct % 5) {
        digitalWrite(RedLED, HIGH);
        digitalWrite(BlueLED, HIGH);
        digitalWrite(YellowLED, HIGH);
      }
      noAct++;
      setLED = millis();
      digitalWrite(Relay, LOW);
      noTone(Buzzer);
    }
  } else if (Mod == ACCESS_GRANTED) {
    digitalWrite(RedLED, LOW);
    digitalWrite(BlueLED, HIGH);
    digitalWrite(YellowLED, LOW);
    digitalWrite(Relay, HIGH);
  } else if (Mod == ACCESS_DENIDED) {
    digitalWrite(RedLED, HIGH);
    digitalWrite(BlueLED, LOW);
    digitalWrite(YellowLED, LOW);
    digitalWrite(Relay, LOW);
  } else if (Mod == CARD_ADDED) {
    digitalWrite(RedLED, LOW);
    digitalWrite(BlueLED, HIGH);
    digitalWrite(YellowLED, HIGH);
    digitalWrite(Relay, LOW);
  } else if (Mod == CARD_DELETED) {
    digitalWrite(RedLED, HIGH);
    digitalWrite(BlueLED, LOW);
    digitalWrite(YellowLED, HIGH);
    digitalWrite(Relay, LOW);
  } else if (Mod == NEW_MASTER) {
    digitalWrite(RedLED, HIGH);
    digitalWrite(BlueLED, HIGH);
    digitalWrite(YellowLED, HIGH);
    digitalWrite(Relay, LOW);
  }
}


//Compare to 4 bytes array
bool checkTwo(byte* a, byte* b) {
  for (uint8_t k = 0; k < 4; k++) {  // Loop 4 times
    if (a[k] != b[k]) {              // IF a != b then false, because: one fails, all fail
      return false;
    }
  }
  return true;
}


//Read card ID
uint8_t ReadCardID(byte* readTag) {
  if (!mfrc522.PICC_ReadCardSerial()) {  //Since a PICC placed get Serial and continue
    return 0;
  }
  // We support 4 byte card ID
  Serial.println(F("Scanned card ID:"));
  for (uint8_t i = 0; i < 4; i++) {  //
    readTag[i] = mfrc522.uid.uidByte[i];
    Serial.print(readTag[i], HEX);
  }
  readTag[4] = '\0';
  Serial.println("");
  return 1;
}


// Check to see if the ID passed is the master programing card
bool isMaster(byte test[]) {
  return checkTwo(test, masterCard);
}


// Read an ID from EEPROM
void readID(uint8_t number) {
  uint8_t start = (number * 4) + 6;          // Figure out starting position
  for (uint8_t i = 0; i < 4; i++) {          // Loop 4 times to get the 4 Bytes
    storedCard[i] = EEPROM.read(start + i);  // Assign values read from EEPROM to array
    delay(1);
  }
}


// Find card ID in EEPROM
bool findID() {
  uint8_t k;
  uint8_t count = EEPROM.read(0);          // Read the first Byte of EEPROM that
  for (uint8_t i = 0; i < count; i++) {    // Loop once for each EEPROM entry
    readID(i);                             // Read an ID from EEPROM, it is stored in storedID
    for (k = 0; k < 4; k++) {              // Loop 4 times
      if (readCard[k] != storedCard[k]) {  // IF a != b then false, because: one fails, all fail
        break;
      }
    }
    if (k == 4) return true;
  }
  // If not, return false
  return false;
}


// Find Slot
uint8_t findIDSLOT() {
  uint8_t k;
  uint8_t count = EEPROM.read(0);          // Read the first Byte of EEPROM that
  for (uint8_t i = 0; i < count; i++) {    // Loop once for each EEPROM entry
    readID(i);                             // Read an ID from EEPROM, it is stored in storedCard[]
    for (k = 0; k < 4; k++) {              // Loop 4 times
      if (readCard[k] != storedCard[k]) {  // IF a != b then false, because: one fails, all fail
        break;
      }
    }
    if (k == 4) return (i);
  }
  return (0);
}


// Remove ID from EEPROM
byte deleteID() {
  uint8_t num = EEPROM.read(0);  // Get the number of used spaces, position 0 stores the number of ID cards
  uint8_t slot;                  // Figure out the slot number of the card
  uint8_t start;                 // = ( num * 4 ) + 6; // Figure out where the next slot starts
  uint8_t looping;               // The number of times the loop repeats
  uint8_t j;
  uint8_t count = EEPROM.read(0);  // Read the first Byte of EEPROM that stores number of cards

  slot = findIDSLOT();  // Figure out the slot number of the card to delete
  start = (slot * 4) + 6;
  looping = ((num - slot) * 4);
  num--;                                                  // Decrement the counter by one
  EEPROM.write(0, num);                                   // Write the new count to the counter
  for (j = 0; j < looping; j++) {                         // Loop the card shift times
    EEPROM.write(start + j, EEPROM.read(start + 4 + j));  // Shift the array values to 4 places earlier in the EEPROM
  }
  for (uint8_t k = 0; k < 4; k++) {  // Shifting loop. Write zero to the last four character in EEPROM
    EEPROM.write(start + j + k, 0);
  }
  Serial.println(F("Succesfully removed ID record from EEPROM"));
  return (1);
}


// Add tag(card) ID to EEPROM
byte writeID() {
  uint8_t num = EEPROM.read(0);            // Get the numer of used spaces, position 0 stores the number of ID cards
  uint8_t start = (num * 4) + 6;           // Figure out where the next slot starts
  num++;                                   // Increment the counter by one
  EEPROM.write(0, num);                    // Write the new count to the counter
  for (uint8_t j = 0; j < 4; j++) {        // Loop 4 times
    EEPROM.write(start + j, readCard[j]);  // Write the array values to EEPROM in the right position
  }
  Serial.println(F("Succesfully added ID record to EEPROM"));
  return (1);
}


// Writing New Master ID to EEPROM
byte writeNewMasterID(char* newMasterCard) {
  uint8_t start = 2;                            // Figure out where the next slot starts
  for (uint8_t j = 0; j < 4; j++) {             // Loop 4 times
    EEPROM.write(start + j, newMasterCard[j]);  // Write the array values to EEPROM in the right position
  }
  Serial.println(F("Succesfully writed New Master ID record to EEPROM"));
  return (1);
}


// Reading Master ID from EEPROM
void readMasterID(byte* masterTag) {
  uint8_t start = 2;                        // Figure out starting position
  for (uint8_t i = 0; i < 4; i++) {         // Loop 4 times to get the 4 Bytes
    masterTag[i] = EEPROM.read(start + i);  // Assign values read from EEPROM to array
    Serial.print(masterTag[i], HEX);
    Serial.print(" ");
    delay(1);
  }
  Serial.println();
  Serial.println(F("Succesfully read Master ID record from EEPROM"));
}


//Delete all tags ID from EEPROM
void deleteEEPROM() {
  digitalWrite(RedLED, HIGH);  // Red Led stays on to inform user we are going to wipe
  Serial.println(F("Wipe Button Pressed"));
  Serial.println(F("You have 10 seconds to Cancel"));
  Serial.println(F("This will be remove all records and cannot be undone"));
  bool buttonState = monitorWipeButton(10000);             // Give user enough time to cancel operation
  if (buttonState == true && digitalRead(wipeB) == LOW) {  // If button still be pressed, wipe EEPROM
    Serial.println(F("Starting Wiping EEPROM"));
    for (uint16_t x = 0; x < EEPROM.length(); x = x + 1) {  //Loop end of EEPROM address
      if (EEPROM.read(x) == 0) {                            //If EEPROM address 0
        // do nothing, already clear, go to the next address in order to save time and reduce writes to EEPROM
      } else {
        EEPROM.write(x, 0);  // if not write 0 to clear, it takes 3.3mS
      }
    }
    Serial.println(F("EEPROM Successfully Wiped"));
    digitalWrite(RedLED, LOW);  // visualize a successful wipe
    delay(200);
    digitalWrite(RedLED, HIGH);
    delay(200);
    digitalWrite(RedLED, LOW);
    delay(200);
    digitalWrite(RedLED, HIGH);
    delay(200);
    digitalWrite(RedLED, LOW);
  } else {
    Serial.println(F("Wiping Cancelled"));  // Show some feedback that the wipe button did not pressed for 15 seconds
    digitalWrite(RedLED, LOW);
  }
}


void setup() {
  //Konfigurisanje pinova
  pinMode(RedLED, OUTPUT);
  pinMode(YellowLED, OUTPUT);
  pinMode(BlueLED, OUTPUT);
  pinMode(Relay, OUTPUT);
  pinMode(Taster, INPUT_PULLUP);
  pinMode(Buzzer, OUTPUT);

  Serial.begin(9600);  // Initiate a serial communication
  SPI.begin();         // Initiate  SPI bus
  mfrc522.PCD_Init();  // Initiate MFRC522

  // Read the Master tag(card) ID from EEPROM
  Serial.println();
  readMasterID(masterCard);
  Serial.println();

  //Inicijalizacija nacina signalizacije
  SigMod = NO_ACTION;
  newCard = true;

  //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  //Ovaj dio koda zamijeniti kodom koji prevodi čitač u MASTER mod
  //ukoliko ID kod MASTER identifikatora nije upisan u EEPROM mikrokontrolera (očitane sve 0 ili sve FF)
  {
    //Wipe taster chack - If the taster (wipeB) pressed longer than 10s it wipes EEPROM
    if (digitalRead(wipeB) == LOW) {  // when button pressed pin should get low, button connected to ground
      newMaster = true;
      if (newMaster) SigMod = NEW_MASTER;
      deleteEEPROM();
      Serial.println(F("New MASTER mode started!!"));
      Serial.println(F("-----------------------------"));
    }

    wipeTaster = digitalRead(wipeB);
    pwipeTaster = wipeTaster;
  }

  //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  //Inicirati promjenljive kojima se prati
  //prisustvo master kartice na čitaču

  // Read the first Byte of EEPROM that
  Serial.print(F("I have "));  // stores the number of ID's in EEPROM
  Serial.print(EEPROM.read(0));
  Serial.print(F(" record(s) on EEPROM"));
  Serial.println("");

  Serial.println(F("Scan PICC..."));
  Serial.println();
}


//Main loop
void loop() 
{
  // Look for new cards
  if (mfrc522.PICC_IsNewCardPresent()) {
    detectionTime = millis();
    successRead = ReadCardID(readCard);
    if (!successRead) return;

    //Ovaj dio koda ne odrađivati ukoliko je detektovano prisustvo
    //master kartice na čitaču
    {
      if (newMaster) {
        if (!findID()) {  // If scanned card is unknown it will become MASTER
          Serial.println(F("Writing new Master ID"));
          Serial.println(F("-----------------------------"));
          writeNewMasterID(readCard);
          //Read the Master tag(card) ID from EEPROM
          readMasterID(masterCard);
          newMaster = false;
          notes = FULL_MELODY;
          melody = AGmelody, noteDur = AGnoteDurations;
          currentNote = 0;
        } else {
          Serial.println(F("Tag/card exists in EEPROM. Try next!"));
          Serial.println(F("-----------------------------"));
          SigMod = NEW_MASTER;
        }
      } else if (programMode) {
        SigMod = PROGRAM_MODE;
        if (isMaster(readCard)) {  //When in program mode check First If master card scanned again to exit program mode
          Serial.println(F("Master Card Scanned"));
          Serial.println(F("Exiting Program Mode"));
          Serial.println(F("-----------------------------"));
          programMode = false;
          SigMod = NO_ACTION;
          //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
          //Ažurirati promjenljive za praćenje prisustva master kartice
        } else {
          if (findID()) {  // If scanned card is known delete it
            Serial.println(F("I know this tag/card, removing..."));
            deleteID();
            Serial.println("-----------------------------");
            Serial.println(F("Scan a tag/card to ADD or REMOVE to EEPROM"));
            SigMod = CARD_DELETED;
            notes = FULL_MELODY;
            melody = ADmelody, noteDur = ADnoteDurations;
            currentNote = 0;
          } else {  // If scanned card is not known add it
            Serial.println(F("I do not know this tag/card, adding..."));
            writeID();
            Serial.println(F("-----------------------------"));
            Serial.println(F("Scan a tag/card to ADD or REMOVE to EEPROM"));
            SigMod = CARD_ADDED;
            notes = FULL_MELODY;
            melody = AGmelody, noteDur = AGnoteDurations;
            currentNote = 0;
          }
        }
      } else {
        if (isMaster(readCard)) {  // If scanned card's ID matches Master Card's ID - enter program mode
          programMode = true;
          Serial.println(F("Hello Master - Entered Program Mode"));
          uint8_t count = EEPROM.read(0);  // Read the first Byte of EEPROM that
          Serial.print(F("I have "));      // stores the number of ID's in EEPROM
          Serial.print(count);
          Serial.print(F(" record(s) on EEPROM"));
          Serial.println("");
          Serial.println(F("Scan a tag(card) to ADD or REMOVE to EEPROM"));
          Serial.println(F("Scan Master tag(card) again to Exit Program Mode"));
          Serial.println(F("-----------------------------"));
          SigMod = PROGRAM_MODE;
          //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
          //Ažurirati promjenljive za praćenje prisustva master kartice
        } else {
          if (findID()) {  // If not, see if the card is in the EEPROM
            Serial.println(F("Welcome, You shall pass"));
            SigMod = ACCESS_GRANTED;
            notes = FULL_MELODY;
            melody = AGmelody, noteDur = AGnoteDurations;
            currentNote = 0;
          } else {  // If not, show that the ID was not valid
            Serial.println(F("You shall not pass"));
            SigMod = ACCESS_DENIDED;
            notes = FULL_MELODY;
            melody = ADmelody, noteDur = ADnoteDurations;
            currentNote = 0;
          }
        }
      }
    }
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    //Karticu stopirati ukoliko nije u pitanju master kartica
    {
      mfrc522.PICC_HaltA();
    }
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    //Ukoliko je u pitanju master izvršiti ponovnu inicijalizaciju čitača
    {

    }
  }
  //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  //Ukoliko nije detektovana kartica na čitaču
  //ažurirati promjenljive koje prate pristustvo master kartice (resetovati njihove vrijednosti)
  {

  }

  //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  //Ovaj dio koda zamijeniti s kodom kojim se provjerava da li je master kartica prisutna na čitaču 
  //10 sekundi i ako jeste izbrisati EEPROM, startovati master mod
  //i inicirati sviranje polovine melodije odbijanja
  { 
    //Wipe taster chack - If the taster (wipeB) pressed longer than 10s it wipes EEPROM
    wipeTaster = digitalRead(wipeB);
    if ((wipeTaster == LOW) && (wipeTaster != pwipeTaster)) {
      newMaster = !newMaster;
      deleteEEPROM();
      if (newMaster) {
        Serial.println(F("New MASTER mode started!!"));
        Serial.println(F("-----------------------------"));
        SigMod = NEW_MASTER;
      }
    }
    pwipeTaster = wipeTaster;
  }

  //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  //Ovdje umetnuti kod kojim se provjerava da li je master kartica prisutna na čitaču 
  //4 sekunde i ako jeste startovati master mod
  //i inicirati sviranje polovine melodije odobrenja
  {

  }


  //Signalisation
  if (millis() - detectionTime > SIG_DURATION) {
    if (newMaster) SigMod = NEW_MASTER;
    else if (programMode) SigMod = PROGRAM_MODE;
    else SigMod = NO_ACTION;
  }
  Signals(SigMod);

  //Zvučni signal
  if (currentNote < notes) {
    Sound(melody, noteDur, currentNote);
    currentNote++;
  } else noTone(Buzzer);
}



//Monitoring of the Wipe button
bool monitorWipeButton(uint32_t interval) {
  uint32_t now = (uint32_t)millis();
  while ((uint32_t)millis() - now < interval) {
    // check on every half a second
    if (((uint32_t)millis() % 500) == 0) {
      if (digitalRead(wipeB) != LOW)
        return false;
    }
  }
  return true;
}
